# -*- coding: utf-8 -*-
'''

  product:  privacyIDEA
  module:   privacyidea controller
    
  Copyright (c) 2014 Cornelius Kölbel
  License:  AGPLv3
  contact:  http://www.privacyidea.org
            
  Description:  This file is part of the privacyidea service
                 
  Dependencies: -
  

'''
import traceback

from pylons import request, response, tmpl_context as c
from pylons.controllers.util import abort, redirect

from privacyidea.lib.base import BaseController
from pylons.templating import render_mako as render

from privacyidea.lib.reply   import sendError
from privacyidea.model.meta  import Session

from privacyidea.lib.util    import get_version
from privacyidea.lib.util    import get_copyright_info
from privacyidea.lib.account import is_admin_identity

from privacyidea.lib.realm    import getRealms
from privacyidea.lib.realm    import getDefaultRealm
from privacyidea.lib.user     import getRealmBox
from pylons import url
from pylons.i18n.translation import _


import logging
import webob


log = logging.getLogger(__name__)


optional = True
required = False

# The HTTP status code, that determines that
# the Login to the selfservice portal is required.
# Is also defined in selfservice.js
LOGIN_CODE = 576

class AccountController(BaseController):
    '''
    The AccountController
        /account/
    is responsible for authenticating the users for the selfservice portal.
    It has the following functions:
        /account/login
        /account/dologin
    '''


    def __before__(self, action, **params):

        log.debug("[__before__::%r] %r" % (action, params))

        try:
            self.set_language()
            c.version = get_version()
            c.licenseinfo = get_copyright_info()

        except webob.exc.HTTPUnauthorized as acc:
            ## the exception, when an abort() is called if forwarded
            log.error("[__before__::%r] webob.exception %r" % (action, acc))
            log.error("[__before__] %s" % traceback.format_exc())
            Session.rollback()
            Session.close()
            raise acc

        except Exception as exx:
            log.error("[__before__::%r] exception %r" % (action, exx))
            log.error("[__before__] %s" % traceback.format_exc())
            Session.rollback()
            Session.close()
            return sendError(response, exx, context='before')

        finally:
            log.debug("[__before__::%r] done" % (action))

    def login(self):
        log.debug("[login] selfservice login screen")
        identity = request.environ.get('repoze.who.identity')
        if identity is not None:
            # After login we either redirect to the manage interface or selfservice
            if is_admin_identity(identity, exception=False):
                redirect("/manage/")
            else:
                redirect("/")
            

                


        res = {}
        try:
            c.defaultRealm = getDefaultRealm()
            res = getRealms()

            c.realmArray = []
            #log.debug("[login] %s" % str(res) )
            for (k, _v) in res.items():
                c.realmArray.append(k)

            c.realmbox = getRealmBox()
            log.debug("[login] displaying realmbox: %i" % int(c.realmbox))

            #TODO: How can we distinguish between failed login and first arrival.
            #c.status = _("Wrong credentials.")

            Session.commit()
            response.status = '%i Logout from privacyIDEA selfservice' % LOGIN_CODE
            return render('/selfservice/login.mako')

        except Exception as e:
            log.error('[login] failed %r' % e)
            log.error('[login] %s' % traceback.format_exc())
            Session.rollback()
            return sendError(response, e)

        finally:
            Session.close()


    def test(self):
        identity = request.environ.get('repoze.who.identity')
        if identity is None:
            # Force skip the StatusCodeRedirect middleware; it was stripping
            #   the WWW-Authenticate header from the 401 response
            request.environ['pylons.status_code_redirect'] = True
            # Return a 401 (Unauthorized) response and signal the repoze.who
            #   basicauth plugin to set the WWW-Authenticate header.
            abort(401, 'You are not authenticated')

        log.debug(u"[test] identity: %r" % identity)
        return """
<body>
Hello, you are logged in as %s.
<a href="/account/logout">logout</a>
</body>
</html>
""" % identity['repoze.who.userid']


#eof##########################################################################

