import logging
log = logging.getLogger(__name__)

import re

from privacyidea.lib.user import getRealmBox
from privacyidea.lib.realm import getDefaultRealm
from privacyidea.lib.selftest import isSelfTest
import traceback
from privacyidea.lib.account import check_admin_password
from privacyidea.lib.account import is_admin_identity
from privacyidea.lib.user import check_user_password

ENCODING = "utf-8"


class UserModelPlugin(object):

    def authenticate(self, environ, identity):
        log.info("[authenticate] entering repoze authenticate function.")
        #log.debug( identity )
        username = None
        realm = None
        success = None
        try:
            if isSelfTest():
                if identity.has_key('login') == False and identity.has_key('repoze.who.plugins.auth_tkt.userid') == True:
                    u = identity.get('repoze.who.plugins.auth_tkt.userid')
                    identity['login'] = u
                    identity['password'] = u

            if getRealmBox():
                username = identity['login']
                realm = identity['realm']
            else:
                log.info("[authenticate] no realmbox, so we are trying to split the loginname")
                m = re.match("(.*)\@(.*)", identity['login'])
                if m:
                    if 2 == len(m.groups()):
                        username = m.groups()[0]
                        realm = m.groups()[1]
                        log.info("[authenticate] found @: username: %r, realm: %r" % (username, realm))
                else:
                    username = identity['login']
                    realm = getDefaultRealm()
                    log.info("[authenticate] using default Realm: username: %r, realm: %r" % (username, realm))

            password = identity['password']
        except KeyError as e:
            log.error("[authenticate] Keyerror in identity: %r." % e)
            log.error("[authenticate] %s" % traceback.format_exc())
            return None

        # check username/realm, password
        if isSelfTest():
            success = "%s@%s" % (unicode(username), unicode(realm))
        else:
            success = check_user_password(username, realm, password)

        if not success and is_admin_identity("%s@%s" % (username, realm), exception=False):
            # user not found or authenticated in resolver. 
            # So let's see, if this is an administrative user.
            success = check_admin_password(username, password, realm)

        return success

    def add_metadata(self, environ, identity):
        #username = identity.get('repoze.who.userid')
        #user = User.get(username)
        #user = "conelius koelbel"
        #log.info( "add_metadata: %s" % identity )

        #pp = pprint.PrettyPrinter(indent=4)
        #log.info("add_meta: environ %s" % pp.pformat(environ)
        log.debug("[add_metadata] add some metatata")
        #for k in environ.keys():
        #    log.debug("add_metadata: environ[%s]: %s" % ( k, environ[k] ))

        for k in identity.keys():
            log.debug("[add_metadata] identity[%s]: %s" % (k, identity[k]))

        #if identity.has_key('realm'):
        #    identity.update( { 'realm' : identity['realm'] } )
        #    log.info("add_metadata: added realm: %s" % identity['realm'] )

        return identity
